@@ -42,7 +42,7 @@
 #pragma warning(pop)
 #endif
 
-#if defined(__FreeBSD__) || defined(__NetBSD__)
+#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__DragonFly__)
 #include <netinet/in.h>
 #endif
 
@@ -323,6 +323,7 @@ struct MEGA_API HttpReq
     bool protect; // check pinned public key
     bool minspeed;
     bool mExpectRedirect = false;
+    bool mChunked = false;
 
     bool sslcheckfailed;
     string sslfakeissuer;
@@ -341,6 +342,9 @@ struct MEGA_API HttpReq
     byte* buf;
     m_off_t buflen, bufpos, notifiedbufpos;
 
+    // When did a post() start
+    std::chrono::steady_clock::time_point postStartTime;
+
     // we assume that API responses are smaller than 4 GB
     m_off_t contentlength;
 
@@ -382,8 +386,8 @@ struct MEGA_API HttpReq
     // a buffer that the HttpReq filled in.   This struct owns the buffer (so HttpReq no longer has it).
     struct http_buf_t
     {
-        byte* datastart();
-        size_t datalen();
+        byte* datastart() const;
+        size_t datalen() const;
 
         size_t start;
         size_t end;
@@ -391,7 +395,7 @@ struct MEGA_API HttpReq
         http_buf_t(byte* b, size_t s, size_t e);  // takes ownership of the byte*, which must have been allocated with new[]
         ~http_buf_t();
         void swap(http_buf_t& other);
-        bool isNull();
+        bool isNull() const;
 
     private:
         byte* buf;
