@@ -101,7 +101,14 @@
 
 // define MEGA_QT_LOGGING to support QString
 #ifdef MEGA_QT_LOGGING
-    #include <QString>
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wconversion"
+#endif
+#include <QString>
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
 #endif
 
 #include "mega/utils.h"
@@ -179,11 +186,6 @@ public:
 
 class SimpleLogger
 {
-public:
-    // flag to turn off logging on the log-output thread, to prevent possible deadlock cycles.
-    static thread_local bool mThreadLocalLoggingDisabled;
-
-private:
     enum LogLevel level;
 
 #ifndef ENABLE_LOG_PERFORMANCE
@@ -193,13 +195,18 @@ private:
 
     std::string getTime();
 #else
-
-#ifdef WIN32
     static thread_local std::array<char, LOGGER_CHUNKS_SIZE> mBuffer;
-#else
-    static __thread std::array<char, LOGGER_CHUNKS_SIZE> mBuffer;
-#endif
     std::array<char, LOGGER_CHUNKS_SIZE>::iterator mBufferIt;
+#ifndef NDEBUG
+    // Detect and warn against multiple instances of this class created in the same thread.
+    // If multiple instances are used in the same thread, the messages from the last created one will
+    // overwrite and corrupt the messages of others, by overwriting mBuffer.
+    //
+    // An alternative approach aiming to allow such cases, would be to overload operator,(), but that will
+    // require to no longer use LoggerVoidify(), and create a "null logger" for the cases when requested
+    // log level needs to be ignored. That would also incur a small performance penalty for the latter case.
+    static thread_local const SimpleLogger* mBufferOwner;
+#endif
 
     using DiffType = std::array<char, LOGGER_CHUNKS_SIZE>::difference_type;
     using NumBuf = char[24];
@@ -371,13 +378,16 @@ private:
     }
 #endif
 
-public:
     static Logger *logger;
 
     static enum LogLevel logCurrentLevel;
 
     static long long maxPayloadLogSize; //above this, the msg will be truncated by [ ... ]
 
+public:
+    // flag to turn off logging on the log-output thread, to prevent possible deadlock cycles.
+    static thread_local bool mThreadLocalLoggingDisabled;
+
     SimpleLogger(const enum LogLevel ll, const char* filename, const int line)
     : level{ll}
 #ifdef ENABLE_LOG_PERFORMANCE
@@ -388,7 +398,16 @@ public:
     {
         if (mThreadLocalLoggingDisabled) return;
 
-#ifndef ENABLE_LOG_PERFORMANCE
+#ifdef ENABLE_LOG_PERFORMANCE
+#ifndef NDEBUG
+        // Multiple instances in the same thread will lead to message corruption!
+        if (!mBufferOwner)
+        {
+            mBufferOwner = this;
+        }
+        assert(mBufferOwner == this);
+#endif
+#else
         if (!logger)
         {
             return;
@@ -442,6 +461,13 @@ public:
         {
             delete s;
         }
+
+#ifndef NDEBUG
+        if (mBufferOwner == this)
+        {
+            mBufferOwner = nullptr;
+        }
+#endif
 #else
         if (logger)
             logger->log(t.c_str(), level, fname.c_str(), ostr.str().c_str());
@@ -605,7 +631,11 @@ public:
     // set the current log level. all logs which are higher than this level won't be handled
     static void setLogLevel(enum LogLevel ll)
     {
-        SimpleLogger::logCurrentLevel = ll;
+        logCurrentLevel = ll;
+    }
+    inline static const LogLevel& getLogLevel()
+    {
+        return logCurrentLevel;
     }
 
     // set the limit of size to requests payload
@@ -613,11 +643,27 @@ public:
     {
         maxPayloadLogSize = size;
     }
+    inline static const long long& getMaxPayloadLogSize()
+    {
+        return maxPayloadLogSize;
+    }
 
     // Log messages forwarded from the client app though the configured logging mechanisms.
     // These do not go through the LOG_<level> macros.
+    //
+    // When ENABLE_LOG_PERFORMANCE is on, this must not be called during the lifetime of an
+    // existing instance created in the same thread. Otherwise this will overwrite the message
+    // of the existing instance. That would be a rare case, but could be achieved by writing
+    // LOG_info << "foo", SimpleLogger::postLog(logInfo, "bar", filename, line);
     static void postLog(LogLevel logLevel, const char *message, const char *filename, int line)
     {
+#ifdef ENABLE_LOG_PERFORMANCE
+#ifndef NDEBUG
+        assert(!mBufferOwner);
+        if (mBufferOwner) return;
+#endif
+#endif
+
         if (logCurrentLevel < logLevel) return;
         SimpleLogger logger(logLevel, filename ? filename : "", line);
         if (message) logger << message;
@@ -637,25 +683,31 @@ template<std::size_t N> inline const char* log_file_leafname( const char (&fullp
 std::ostream& operator <<(std::ostream&, const std::system_error&);
 std::ostream& operator <<(std::ostream&, const std::error_code&);
 
+// Helper used in LOG_* macros below to make the right operand of ?: void to match the left one
+struct LoggerVoidify
+{
+    void operator&(SimpleLogger&) {}
+};
+
 #define LOG_verbose \
-    if (::mega::SimpleLogger::logCurrentLevel >= ::mega::logMax) \
-        ::mega::SimpleLogger(::mega::logMax, ::mega::log_file_leafname(__FILE__), __LINE__)
+    ::mega::SimpleLogger::getLogLevel() < ::mega::logMax ? (void)0 : \
+        ::mega::LoggerVoidify() & ::mega::SimpleLogger(::mega::logMax, ::mega::log_file_leafname(__FILE__), __LINE__)
 
 #define LOG_debug \
-    if (::mega::SimpleLogger::logCurrentLevel >= ::mega::logDebug) \
-        ::mega::SimpleLogger(::mega::logDebug, ::mega::log_file_leafname(__FILE__), __LINE__)
+    ::mega::SimpleLogger::getLogLevel() < ::mega::logDebug ? (void)0 : \
+        ::mega::LoggerVoidify() & ::mega::SimpleLogger(::mega::logDebug, ::mega::log_file_leafname(__FILE__), __LINE__)
 
 #define LOG_info \
-    if (::mega::SimpleLogger::logCurrentLevel >= ::mega::logInfo) \
-        ::mega::SimpleLogger(::mega::logInfo, ::mega::log_file_leafname(__FILE__), __LINE__)
+    ::mega::SimpleLogger::getLogLevel() < ::mega::logInfo ? (void)0 : \
+        ::mega::LoggerVoidify() & ::mega::SimpleLogger(::mega::logInfo, ::mega::log_file_leafname(__FILE__), __LINE__)
 
 #define LOG_warn \
-    if (::mega::SimpleLogger::logCurrentLevel >= ::mega::logWarning) \
-        ::mega::SimpleLogger(::mega::logWarning, ::mega::log_file_leafname(__FILE__), __LINE__)
+    ::mega::SimpleLogger::getLogLevel() < ::mega::logWarning ? (void)0 : \
+        ::mega::LoggerVoidify() & ::mega::SimpleLogger(::mega::logWarning, ::mega::log_file_leafname(__FILE__), __LINE__)
 
 #define LOG_err \
-    if (::mega::SimpleLogger::logCurrentLevel >= ::mega::logError) \
-        ::mega::SimpleLogger(::mega::logError, ::mega::log_file_leafname(__FILE__), __LINE__)
+    ::mega::SimpleLogger::getLogLevel() < ::mega::logError ? (void)0 : \
+        ::mega::LoggerVoidify() & ::mega::SimpleLogger(::mega::logError, ::mega::log_file_leafname(__FILE__), __LINE__)
 
 #define LOG_fatal \
     ::mega::SimpleLogger(::mega::logFatal, ::mega::log_file_leafname(__FILE__), __LINE__)
@@ -693,11 +745,20 @@ public:
 #endif
     ) override;
 
+    // Do not use this unless you know what you are doing!
+    //
+    // This is an unfortunate workaround for cases when multiple connections/clients add
+    // each its own logger to the same target (i.e. file), leading to duplicated messages
+    // being logged consecutively.
+    // This for example has happened in MegaChat's automated tests.
+    void useOnlyFirstLogger(bool onlyFirst = true) { useOnlyFirstMegaLogger = onlyFirst; }
+
 private:
     std::recursive_mutex mutex;
     map<void*, LogCallback> megaLoggers;
     bool logToConsole;
     bool alreadyLogging = false;
+    bool useOnlyFirstMegaLogger = false;
 };
 
 
