@@ -257,10 +257,8 @@ struct FileAttributesPending : public mapWithLookupExisting<UploadHandle, Upload
     }
 };
 
-
 class MegaClient;
 
-
 class MEGA_API KeyManager
 {
 public:
@@ -296,6 +294,8 @@ public:
     bool addShareKey(handle sharehandle, std::string shareKey, bool sharedSecurely = false);
     string getShareKey(handle sharehandle) const;
     bool isShareKeyTrusted(handle sharehandle) const;
+    bool isShareKeyInUse(handle sharehandle) const;
+    void setSharekeyInUse(handle sharehandle, bool sent);
 
     // return empty string if the user's credentials are not verified (or if fail to encrypt)
     std::string encryptShareKeyTo(handle userhandle, std::string shareKey);
@@ -311,7 +311,6 @@ public:
     bool isUnverifiedOutShare(handle nodeHandle, const string& uid);
     bool isUnverifiedInShare(handle nodeHandle, handle userHandle);
 
-    void cacheShareKeys();
     void loadShareKeys();
 
     void commit(std::function<void()> applyChanges, std::function<void()> completion = nullptr);
@@ -320,6 +319,12 @@ public:
     // returns a formatted string, for logging purposes
     string toString() const;
 
+    // Returns true if the warnings related to shares with non-verified contacts are enabled.
+    bool getContactVerificationWarning();
+
+    // Enable/disable the warnings for shares with non-verified contacts.
+    void setContactVerificationWarning(bool enabled);
+
     // this method allows to change the feature-flag for testing purposes
     void setSecureFlag(bool enabled) { mSecure = enabled; }
 
@@ -355,6 +360,17 @@ private:
         TAG_WARNINGS = 96,
     };
 
+    // Bit position for different flags for each sharekey. Bits 2 to 7 reserved for future usage.
+    enum ShareKeyFlagsId
+    {
+        TRUSTED = 0,    // If the sharekey is trusted
+        INUSE = 1,      // If there is an active outshare or folder-link using the sharekey
+    };
+
+    // Bitmap with flags for each sharekey. The field is 1 byte size in the attribute.
+    // See used bits and flag meaning in "ShareKeyFlagsId" enumeration.
+    typedef std::bitset<8> ShareKeyFlags;
+
     static const uint8_t IV_LEN = 12;
     static const std::string SVCRYPTO_PAIRWISE_KEY;
 
@@ -388,11 +404,10 @@ private:
     string mPrivEd25519, mPrivCu25519, mPrivRSA;
     string mAuthEd25519, mAuthCu25519;
     string mBackups;
-    string mWarnings;
     string mOther;
 
-    // maps node handle of the shared folder to a pair of sharekey bytes and trust flag
-    map<handle, pair<string, bool>> mShareKeys;
+    // maps node handle of the shared folder to a pair of sharekey bytes and sharekey flags.
+    map<handle, pair<string, ShareKeyFlags>> mShareKeys;
 
     // maps node handle to the target users (where value can be a user's handle in B64 or the email address)
     map<handle, set<string>> mPendingOutShares;
@@ -400,6 +415,9 @@ private:
     // maps base64 node handles to pairs of source user handle and share key
     map<string, pair<handle, string>> mPendingInShares;
 
+    // warnings as stored as a key-value map
+    map<string, string> mWarnings;
+
     // decode data from the decrypted ^!keys attribute and stores values in `km`
     // returns false in case of unserializatison isues
     static bool unserialize(KeyManager& km, const string& keysContainer);
@@ -407,6 +425,11 @@ private:
     // prepares the header for a new serialized record of type 'tag' and 'len' bytes
     string tagHeader(const byte tag, size_t len) const;
 
+    // Serialize pairs of tags and values as Length+Tag+Lengh+Value.
+    // warnings and pending inshares are encoded like that when serialized.
+    static bool deserializeFromLTLV(const string& blob, map<string, string>& data);
+    static string serializeToLTLV(const map<string, string>& data);
+
     // encode data from the decrypted ^!keys attribute
     string serialize() const;
 
@@ -425,6 +448,10 @@ private:
     string serializeBackups() const;
     static bool deserializeBackups(KeyManager& km, const string& blob);
 
+    string serializeWarnings() const;
+    static bool deserializeWarnings(KeyManager& km, const string& blob);
+    static string warningsToString(const KeyManager& km);
+
     std::string computeSymmetricKey(handle user);
 
     // validates data in `km`: ie. downgrade attack, tampered keys...
@@ -440,7 +467,7 @@ private:
     void updateAuthring(attr_t at, std::string &value);
 
     // update sharekeys (incl. trust). It doesn't purge non-existing items
-    void updateShareKeys(map<handle, pair<std::string, bool> > &shareKeys);
+    void updateShareKeys(map<handle, pair<std::string, ShareKeyFlags> > &shareKeys);
 
     // true if the credentials of this user require verification
     bool verificationRequired(handle userHandle);
@@ -484,12 +511,20 @@ public:
     // Don't start showing the cookie banner until API says so
     bool mCookieBannerEnabled = false;
 
+    // Consider an account as new if it was created less than X days earlier (right now it's 30days; received in "ug":"na")
+    bool accountIsNew = false;
+
+    // AB Test flags
+    std::map<string, uint32_t> mABTestFlags;
+
 private:
     // Pro Flexi plan is enabled
     bool mProFlexi = false;
 public:
     bool isProFlexi() const { return mProFlexi; }
 
+    Error sendABTestActive(const char* flag, CommandABTestActive::Completion completion);
+
     // 2 = Opt-in and unblock SMS allowed 1 = Only unblock SMS allowed 0 = No SMS allowed  -1 = flag was not received
     SmsVerificationState mSmsVerificationState;
 
@@ -502,7 +537,8 @@ public:
     bool ephemeralSession = false;
     bool ephemeralSessionPlusPlus = false;
 
-    static string publicLinkURL(bool newLinkFormat, nodetype_t type, handle ph, const char *key);
+    static TypeOfLink validTypeForPublicURL(nodetype_t type);
+    static string publicLinkURL(bool newLinkFormat, TypeOfLink type, handle ph, const char *key);
 
     string getWritableLinkAuthKey(handle node);
 
@@ -542,7 +578,7 @@ public:
     void cancelsignup();
 
     // full account confirmation/creation support
-    string sendsignuplink2(const char*, const char *, const char*);
+    string sendsignuplink2(const char*, const char *, const char*, int ctag = 0);
     void resendsignuplink2(const char*, const char *);
 
     void confirmsignuplink2(const byte*, unsigned);
@@ -566,6 +602,9 @@ public:
     // session login: binary session, bytecount
     void login(string session);
 
+    // handle login result, and allow further actions when successful
+    void loginResult(error e, std::function<void()> onLoginOk = nullptr);
+
     // check password
     error validatepwd(const char* pswd);
     bool validatepwdlocally(const char* pswd);
@@ -579,13 +618,14 @@ public:
     // get the public key of an user
     void getpubkey(const char* user);
 
-    // check if logged in
+    // check if logged in (avoid repetitive calls <-- requires call to Cryptopp::InverseMod(), which is slow)
     sessiontype_t loggedin();
 
     // provide state by change callback
     void reportLoggedInChanges();
     sessiontype_t mLastLoggedInReportedState = NOTLOGGEDIN;
     handle mLastLoggedInMeHandle = UNDEF;
+    string mLastLoggedInMyEmail;
 
     // check the reason of being blocked
     void whyamiblocked();
@@ -617,7 +657,7 @@ public:
     void killallsessions();
 
     // extract public handle and key from a public file/folder link
-    error parsepubliclink(const char *link, handle &ph, byte *key, bool isFolderLink);
+    error parsepubliclink(const char *link, handle &ph, byte *key, TypeOfLink type);
 
     // open the SC database and get the SCSN from it
     void checkForResumeableSCDatabase();
@@ -645,9 +685,6 @@ public:
     // fetchnodes stats
     FetchNodesStats fnstats;
 
-    // load cryptographic keys: RSA, Ed25519, Cu25519 and their signatures
-    void fetchkeys();
-
     // check existence and integrity of keys and signatures, initialize if missing
     void initializekeys();
 
@@ -670,10 +707,10 @@ public:
     error updateAuthring(AuthRing *authring, attr_t authringType, bool temporalAuthring, handle updateduh);
 
     // set the Ed25519 public key as verified for a given user in the authring (done by user manually by comparing hash of keys)
-    error verifyCredentials(handle uh);
+    error verifyCredentials(handle uh, std::function<void(Error)>);
 
     // reset the authentication method of Ed25519 key from Fingerprint-verified to Seen for a given user
-    error resetCredentials(handle uh);
+    error resetCredentials(handle uh, std::function<void (Error)>);
 
     // check credentials are verified for a given user
     bool areCredentialsVerified(handle uh);
@@ -714,7 +751,7 @@ public:
     void removeOutSharesFromSubtree(Node* n, int tag);
 
     // start/stop/pause file transfer
-    bool startxfer(direction_t, File*, TransferDbCommitter&, bool skipdupes, bool startfirst, bool donotpersist, VersioningOption, error* cause, int tag);
+    bool startxfer(direction_t, File*, TransferDbCommitter&, bool skipdupes, bool startfirst, bool donotpersist, VersioningOption, error* cause, int tag, m_off_t availableDiskSpace = 0);
     void stopxfer(File* f, TransferDbCommitter* committer);
     void pausexfers(direction_t, bool pause, bool hard, TransferDbCommitter& committer);
 
@@ -746,7 +783,7 @@ public:
 
     // helpfer function for preparing a putnodes call for new node
     error putnodes_prepareOneFile(NewNode* newnode, Node* parentNode, const char *utf8Name, const UploadToken& binaryUploadToken,
-                                  byte *theFileKey, char *megafingerprint, const char *fingerprintOriginal,
+                                  const byte *theFileKey, const char *megafingerprint, const char *fingerprintOriginal,
                                   std::function<error(AttrMap&)> addNodeAttrsFunc = nullptr,
                                   std::function<error(std::string *)> addFileAttrsFunc = nullptr);
 
@@ -764,7 +801,7 @@ public:
     void putnodes(const char*, vector<NewNode>&&, int tag, CommandPutNodes::Completion&& completion = nullptr);
 
     // attach file attribute to upload or node handle
-    void putfa(NodeOrUploadHandle, fatype, SymmCipher*, int tag, std::unique_ptr<string>);
+    bool putfa(NodeOrUploadHandle, fatype, SymmCipher*, int tag, std::unique_ptr<string>);
 
     // move as many as possible from pendingfa to activefa
     void activatefa();
@@ -791,6 +828,19 @@ public:
     // retrieve the email address of a user
     void getUserEmail(const char *uid);
 
+
+//
+// Account upgrade to V2
+//
+public:
+    void saveV1Pwd(const char* pwd);
+private:
+    void upgradeAccountToV2(const string& pwd, int ctag, std::function<void(error e)> completion);
+    // temporarily stores v1 account password, to allow automatic upgrade to v2 after successful (full-)login
+    unique_ptr<pair<string, SymmCipher>> mV1PswdVault;
+// -------- end of Account upgrade to V2
+
+public:
 #ifdef DEBUG
     // queue a user attribute removal
     void delua(const char* an);
@@ -805,6 +855,9 @@ public:
     // Migrate the account to start using the new ^!keys attr.
     void upgradeSecurity(std::function<void(Error)> completion);
 
+    // Set the flag to enable/disable warnings when sharing with a non-verified contact.
+    void setContactVerificationWarning(bool enabled, std::function<void(Error)> completion = nullptr);
+
     // Creates a new share key for the node if there is no share key already created.
     void openShareDialog(Node* n, std::function<void (Error)> completion);
 
@@ -895,6 +948,11 @@ public:
     typedef std::function<void(std::function<void()> continuation)> UndoFunction;
     void preparebackup(SyncConfig, std::function<void(Error, SyncConfig, UndoFunction revertOnError)>);
 
+    /**
+     * @brief Used to migrate all the sync's management data from MEGAsync to the SDK
+     *
+     * @deprecated This function is deprecated. Please don't use it in new code.
+     */
     void copySyncConfig(const SyncConfig& config, std::function<void(handle, error)> completion);
 
     /**
@@ -1028,8 +1086,8 @@ public:
     void userfeedbackstore(const char *);
 
     // send event
-    void sendevent(int, const char *);
-    void sendevent(int, const char *, int tag);
+    void sendevent(int, const char *, const char* viewId = nullptr, bool addJourneyId = false);
+    void sendevent(int, const char *, int tag, const char* viewId = nullptr, bool addJourneyId = false);
 
     // create support ticket
     void supportticket(const char *message, int type);
@@ -1063,6 +1121,9 @@ public:
     // get the URL of a chat
     void getUrlChat(handle chatid);
 
+    // set chat mode (public/private)
+    void setChatMode(TextChat* chat, bool pubChat);
+
     // process object arrays by the API server (users + privileges)
     userpriv_vector * readuserpriv(JSON* j);
 
@@ -1109,10 +1170,13 @@ public:
 
     // parse scheduled meeting or scheduled meeting occurrences
     error parseScheduledMeetings(std::vector<std::unique_ptr<ScheduledMeeting> > &schedMeetings,
-                                 bool parsingOccurrences, JSON *j = nullptr, bool parseOnce = false,
+                                 bool parsingOccurrences, JSON *j, bool parseOnce = false,
                                  handle* originatingUser = nullptr,
                                  UserAlert::UpdatedScheduledMeeting::Changeset* cs = nullptr,
                                  handle_set* childMeetingsDeleted = nullptr);
+
+    // report invalid scheduled meeting by sending an event to stats server
+    void reportInvalidSchedMeeting(const ScheduledMeeting* sched = nullptr);
 #endif
 
     // get mega achievements
@@ -1313,7 +1377,6 @@ private:
     // Working lock
     unique_ptr<HttpReq> workinglockcs;
 
-private:
     // Request status monitor
     unique_ptr<HttpReq> mReqStatCS;
 
@@ -1415,12 +1478,15 @@ public:
     void createUpdatedSMAlert(const handle&, handle chatid, handle schedId, handle parentSchedId,
                                m_time_t startDateTime, UserAlert::UpdatedScheduledMeeting::Changeset&& cs);
     static error parseScheduledMeetingChangeset(JSON*, UserAlert::UpdatedScheduledMeeting::Changeset*);
+    void clearSchedOccurrences(TextChat& chat);
 #endif
     void sc_uac();
+    void sc_uec();
     void sc_la();
     void sc_ub();
     void sc_sqac();
     void sc_pk();
+    void sc_cce();
 
     void init();
 
@@ -1465,7 +1531,6 @@ public:
 
     void dodiscarduser(User* u, bool discardnotified);
 
-public:
     void enabletransferresumption(const char *loggedoutid = NULL);
     void disabletransferresumption(const char *loggedoutid = NULL);
 
@@ -1496,6 +1561,8 @@ public:
     // NodeManager instance to wrap all access to Node objects
     NodeManager mNodeManager;
 
+    mutex nodeTreeMutex;
+
     // there is data to commit to the database when possible
     bool pendingsccommit;
 
@@ -1578,9 +1645,6 @@ public:
     // flag to pause / resume the processing of action packets
     bool scpaused;
 
-    // MegaClient-Server response JSON
-    JSON json;
-
     // Server-MegaClient request JSON and processing state flag ("processing a element")
     JSON jsonsc;
     bool insca;
@@ -1695,7 +1759,7 @@ public:
     static dstime USER_DATA_EXPIRATION_BACKOFF_SECS;
 
     // total number of Node objects
-    long long totalNodes;
+    std::atomic_ullong totalNodes;
 
     // tracks how many nodes have had a successful applykey()
     long long mAppliedKeyNodeCount = 0;
@@ -1703,16 +1767,22 @@ public:
     // server-client request sequence number
     SCSN scsn;
 
+    // process an array of users from the API server
     bool readusers(JSON*, bool actionpackets);
 
+    // process a JSON user object
+    // possible results:
+    // 0 -> no object found
+    // 1 -> successful parsing
+    // any other number -> parsing error
+    int readuser(JSON*, bool actionpackets);
+
     user_vector usernotify;
     void notifyuser(User*);
 
     pcr_vector pcrnotify;
     void notifypcr(PendingContactRequest*);
 
-    void notifynode(Node*);
-
     // update transfer in the persistent cache
     void transfercacheadd(Transfer*, TransferDbCommitter*);
 
@@ -1770,6 +1840,21 @@ public:
     // determine if the file is a document.
     bool nodeIsDocument(const Node *n) const;
 
+    // determine if the file is a PDF.
+    bool nodeIsPdf(const Node *n) const;
+
+    // determine if the file is a presentation.
+    bool nodeIsPresentation(const Node *n) const;
+
+    // determine if the file is an archive.
+    bool nodeIsArchive(const Node* n) const;
+
+    // determine if the file is a program.
+    bool nodeIsProgram(const Node* n) const;
+
+    // determine if the file is miscellaneous.
+    bool nodeIsMiscellaneous(const Node* n) const;
+
     // functions for determining whether we can clone a node instead of upload
     // or whether two files are the same so we can just upload/download the data once
     bool treatAsIfFileDataEqual(const FileFingerprint& nodeFingerprint, const LocalPath& file2, const string& filenameExtensionLowercaseNoDot);
@@ -1945,6 +2030,14 @@ public:
     // process object arrays by the API server
     int readnodes(JSON*, int, putsource_t, vector<NewNode>*, bool modifiedByThisClient, bool applykeys);
 
+    // process a JSON node object
+    // possible results:
+    // 0 -> no object found
+    // 1 -> successful parsing
+    // any other number -> parsing error
+    int readnode(JSON*, int, putsource_t, vector<NewNode>*, bool modifiedByThisClient, bool applykeys,
+                 NodeManager::MissingParentNodes& missingParentNodes, handle &previousHandleForAlert);
+
     void readok(JSON*);
     void readokelement(JSON*);
     void readoutshares(JSON*);
@@ -1955,7 +2048,14 @@ public:
 
     error readmiscflags(JSON*);
 
-    void procph(JSON*);
+    bool procph(JSON*);
+
+    // process a JSON ph object
+    // possible results:
+    // 0 -> no object found
+    // 1 -> successful parsing
+    // any other number -> parsing error
+    int procphelement(JSON*);
 
     void procsnk(JSON*);
     void procsuk(JSON*);
@@ -1997,6 +2097,7 @@ public:
     static const int CHATLINKHANDLE = 6;
     static const int SETHANDLE = Set::HANDLESIZE;
     static const int SETELEMENTHANDLE = SetElement::HANDLESIZE;
+    static const int PUBLICSETHANDLE = Set::PUBLICHANDLESIZE;
 
     // max new nodes per request
     static const int MAX_NEWNODES = 2000;
@@ -2038,12 +2139,6 @@ public:
     // Pending contact keys during initialization
     std::map<attr_t, set<handle>> mPendingContactKeys;
 
-    // number of authrings being fetched
-    unsigned short mFetchingAuthrings = 0;
-
-    // actual state of keys
-    bool fetchingkeys;
-
     // invalidate received keys (when fail to load)
     void clearKeys();
 
@@ -2100,6 +2195,12 @@ public:
     // create a new folder with given name and stores its node's handle into the user's attribute ^!bak
     error setbackupfolder(const char* foldername, int tag, std::function<void(Error)> addua_completion);
 
+    // fetch backups and syncs from BC, search bkpId among them, disable the backup or sync, update sds attribute, for a backup move or delete its contents
+    void removeFromBC(handle bkpId, handle bkpDest, std::function<void(const Error&)> f);
+
+    // fetch backups and syncs from BC
+    void getBackupInfo(std::function<void(const Error&, const vector<CommandBackupSyncFetch::Data>&)> f);
+
     // sets the auth token to be used when logged into a folder link
     void setFolderLinkAccountAuth(const char *auth);
 
@@ -2273,9 +2374,6 @@ public:
     MegaClient(MegaApp*, shared_ptr<Waiter>, HttpIO*, DbAccess*, GfxProc*, const char*, const char*, unsigned workerThreadCount);
     ~MegaClient();
 
-    void filenameAnomalyDetected(FilenameAnomalyType type, const LocalPath& localPath, const string& remotePath);
-    unique_ptr<FilenameAnomalyReporter> mFilenameAnomalyReporter;
-
 struct MyAccountData
 {
     void setProLevel(AccountType prolevel) { mProLevel = prolevel; }
@@ -2291,6 +2389,42 @@ private:
     m_time_t mProUntil = -1;
 } mMyAccount;
 
+// JourneyID for cs API requests and log events. Populated from "ug"/"gmf" commands response.
+// It is kept in memory and persisted in disk until a full logout.
+struct JourneyID
+{
+private:
+    // The JourneyID value - a 16-char hex string (or an empty string if it hasn't been retrieved yet)
+    string mJidValue;
+    // The tracking flag: used to attach the JourneyID to cs requests
+    bool mTrackValue;
+    // Local cache file
+    unique_ptr<FileSystemAccess>& mClientFsaccess;
+    LocalPath mCacheFilePath;
+    bool storeValuesToCache(bool storeJidValue, bool storeTrackValue) const;
+
+public:
+    static constexpr size_t HEX_STRING_SIZE = 16;
+    JourneyID(unique_ptr<FileSystemAccess>& clientFsaccess, const LocalPath& rootPath);
+    // Updates the JourneyID and the tracking flag based on the provided jidValue, which must be a 16-char hex string.
+    // When jidValue is not empty:
+    // - Sets mJidValue to jidValue only if it is currently unset (empty).
+    // - Sets mTrackValue if it is currently unset (false).
+    // When jidValue is empty:
+    // - Keeps mJidValue unchanged.
+    // - Unsets mTrackValue if it is currently set (true).
+    // Returns true if either the JourneyID (mJidValue) or the tracking flag (mTrackValue) have been updated.
+    bool setValue(const string& jidValue);
+    // Get the JourneyID (empty if still unset)
+    string getValue() const;
+    // Check if the tracking flag is set, i.e.: the JourneyID must be tracked (used in cs API reqs)
+    bool isTrackingOn() const;
+    // Load the JourneyID and the tracking flag stored in the cache file.
+    bool loadValuesFromCache();
+    // Remove local cache file and reset the JourneyID so a new one can be set from the next "ug"/"gmf" command.
+    bool resetCacheAndValues();
+};
+
 private:
     // Since it's quite expensive to create a SymmCipher, this are provided to use for quick operations - just set the key and use.
     SymmCipher tmpnodecipher;
@@ -2300,15 +2434,41 @@ private:
 
     error changePasswordV1(User* u, const char* password, const char* pin);
     error changePasswordV2(const char* password, const char* pin);
+    void fillCypheredAccountDataV2(const char* password, vector<byte>& clientRandomValue, vector<byte>& encmasterkey,
+                                   string& hashedauthkey, string& salt);
 
     static vector<byte> deriveKey(const char* password, const string& salt, size_t derivedKeySize);
 
+//
+// JourneyID and ViewID
+//
+    // JourneyID for cs API requests and log events
+    JourneyID mJourneyId;
+
+public:
+
+    // Checks if there is a valid JourneyID and tracking flag is set
+    bool trackJourneyId() const;
+
+    // Retrieves the JourneyID value, which is a 16-character hexadecimal string (for submission to the API)
+    // If the JourneyID is still unset, it returns an empty string.
+    string getJourneyId() const;
+
+    // Load the JourneyID values from the local cache.
+    bool loadJourneyIdCacheValues();
+
+    // Set the JourneyID value from a 16-character hexadecimal string (obtained from API commands "ug"/"gmf")
+    // See JourneyID::setValue() for full doc
+    bool setJourneyId(const string& jid);
+
+    // Generates a unique ViewID that the caller should store and can optionally use in subsequent sendevent() calls.
+    // ViewID is employed by apps for event logging. It is generated by the SDK to ensure consistent and shared logic across applications.
+    static string generateViewId(PrnGen& rng);
 
 //
 // Sets and Elements
 //
 
-public:
     // generate "asp" command
     void putSet(Set&& s, std::function<void(Error, const Set*)> completion);
 
@@ -2316,7 +2476,7 @@ public:
     void removeSet(handle sid, std::function<void(Error)> completion);
 
     // generate "aft" command
-    void fetchSet(handle sid, std::function<void(Error, Set*, map<handle, SetElement>*)> completion);
+    void fetchSetInPreviewMode(std::function<void(Error, Set*, elementsmap_t*)> completion);
 
     // generate "aepb" command
     void putSetElements(vector<SetElement>&& els, std::function<void(Error, const vector<const SetElement*>*, const vector<int64_t>*)> completion);
@@ -2331,10 +2491,10 @@ public:
     void removeSetElement(handle sid, handle eid, std::function<void(Error)> completion);
 
     // handle "aesp" parameter, part of 'f'/ "fetch nodes" response
-    bool procaesp();
+    bool procaesp(JSON& j);
 
     // load Sets and Elements from json
-    error readSetsAndElements(JSON& j, map<handle, Set>& newSets, map<handle, map<handle, SetElement>>& newElements);
+    error readSetsAndElements(JSON& j, map<handle, Set>& newSets, map<handle, elementsmap_t>& newElements);
 
     // return Set with given sid or nullptr if it was not found
     const Set* getSet(handle sid) const;
@@ -2358,7 +2518,7 @@ public:
     const SetElement* getSetElement(handle sid, handle eid) const;
 
     // return all available Elements in a Set, indexed by eid
-    const map<handle, SetElement>* getSetElements(handle sid) const;
+    const elementsmap_t* getSetElements(handle sid) const;
 
     // add new SetElement or replace exisiting one
     const SetElement* addOrUpdateSetElement(SetElement&& el);
@@ -2366,12 +2526,41 @@ public:
     // delete Element with eid from Set with sid in local memory; return true if found and deleted
     bool deleteSetElement(handle sid, handle eid);
 
-private:
+    // return true if Set with given sid is exported (has a public link)
+    bool isExportedSet(handle sid) const;
 
+    void exportSet(handle sid, bool makePublic, std::function<void(Error)> completion);
+
+    // returns result of the operation and the link created
+    pair<error, string> getPublicSetLink(handle sid) const;
+
+    // returns error code and public handle for the link provided as a param
+    error fetchPublicSet(const char* publicSetLink, std::function<void(Error, Set*, elementsmap_t*)>);
+
+    void stopSetPreview() { if (mPreviewSet) mPreviewSet.reset(); }
+
+    bool inPublicSetPreview() const { return !!mPreviewSet; }
+
+    const SetElement* getPreviewSetElement(handle eid) const
+    { return isElementInPreviewSet(eid) ? &mPreviewSet->mElements[eid] : nullptr; }
+
+    const Set* getPreviewSet() const { return inPublicSetPreview() ? &mPreviewSet->mSet : nullptr; }
+    const elementsmap_t* getPreviewSetElements() const
+    { return inPublicSetPreview() ? &mPreviewSet->mElements : nullptr; }
+
+private:
     error readSets(JSON& j, map<handle, Set>& sets);
     error readSet(JSON& j, Set& s);
-    error readElements(JSON& j, map<handle, map<handle, SetElement>>& elements);
+    error readElements(JSON& j, map<handle, elementsmap_t>& elements);
     error readElement(JSON& j, SetElement& el);
+    error readAllNodeMetadata(JSON& j, map<handle, SetElement::NodeMetadata>& nodes);
+    error readSingleNodeMetadata(JSON& j, SetElement::NodeMetadata& node);
+    bool decryptNodeMetadata(SetElement::NodeMetadata& nodeMeta, const string& key);
+    error readExportedSet(JSON& j, Set& s, pair<bool, m_off_t>& exportRemoved);
+    error readSetsPublicHandles(JSON& j, map<handle, Set>& sets);
+    error readSetPublicHandle(JSON& j, map<handle, Set>& sets);
+    void fixSetElementWithWrongKey(const Set& set);
+    size_t decryptAllSets(map<handle, Set>& newSets, map<handle, elementsmap_t>& newElements, map<handle, SetElement::NodeMetadata>* nodeData);
     error decryptSetData(Set& s);
     error decryptElementData(SetElement& el, const string& setKey);
     string decryptKey(const string& k, SymmCipher& cipher) const;
@@ -2382,6 +2571,7 @@ private:
     void sc_asr(); // AP after removed Set
     void sc_aep(); // AP after new or updated Set Element
     void sc_aer(); // AP after removed Set Element
+    void sc_ass(); // AP after exported set update
 
     bool initscsets();
     bool fetchscset(string* data, uint32_t id);
@@ -2398,10 +2588,73 @@ private:
     void notifysetelement(SetElement*);
     void clearsetelementnotify(handle sid);
     vector<SetElement*> setelementnotify;
-    map<handle, map<handle, SetElement>> mSetElements; // indexed by Set id, then Element id
+    map<handle, elementsmap_t> mSetElements; // indexed by Set id, then Element id
 
+    struct SetLink
+    {
+        handle mPublicId = UNDEF; // same as mSet.mPublicId once fetched
+        string mPublicKey;
+        string mPublicLink;
+        Set mSet;
+        elementsmap_t mElements;
+    };
+    unique_ptr<SetLink> mPreviewSet;
+
+    bool isElementInPreviewSet(handle eid) const
+    { return mPreviewSet && (mPreviewSet->mElements.find(eid) != end(mPreviewSet->mElements)); }
 // -------- end of Sets and Elements
 
+    // Generates a key pair (x25519 (Cu) key pair) to use for Vpn Credentials (MegaClient::putVpnCredential)
+    StringKeyPair generateVpnKeyPair();
+
+public:
+
+/* Mega VPN methods */
+
+    // Call "vpnr" command.
+    void getVpnRegions(CommandGetVpnRegions::Cb&& = nullptr /* Completion */);
+
+    // Call "vpng" command.
+    void getVpnCredentials(CommandGetVpnCredentials::Cb&& = nullptr /* Completion */);
+
+    // Call "vpnp" command.
+    void putVpnCredential(std::string&& /* VPN Region */, CommandPutVpnCredential::Cb&& = nullptr /* Completion */);
+
+    // Call "vpnd" command.
+    void delVpnCredential(int /* SlotID */, CommandDelVpnCredential::Cb&& = nullptr /* Completion */);
+
+    // Call "vpnc" command.
+    void checkVpnCredential(std::string&& /* User Public Key */, CommandCheckVpnCredential::Cb&& = nullptr /* Completion */);
+
+    /**
+     * @brief Generates a VPN credential string equivalent to the .conf file generated by the webclient.
+     * This method is meant to be called with the credential details obtained from CommandPutVpnCredential.
+     *
+     * Content:
+     *    [Interface]
+     *    PrivateKey = User Private Key
+     *    Address = IPv4, IPv6
+     *    DNS = IPv4, IPv6
+     *
+     *    [Peer]
+     *    PublicKey = Cluster Public Key
+     *    AllowedIPs = 0.0.0.0/0, ::/0
+     *    Endpoint = host:port
+     *
+     * @note These VPN credential details are not the same than the data obtained from MegaClient::getVpnCredentials()
+     * @see MegaClient::putVpnCredential()
+     * @return The string with the VPN credentials specified above.
+    */
+    string generateVpnCredentialString(int /* ClusterID */,
+                                       std::string&& /* VPN Region */,
+                                       std::string&& /* IPv4 */,
+                                       std::string&& /* IPv6 */,
+                                       StringKeyPair&& /* Peer Key Pair <User Private, Cluster Public> */);
+
+/* Mega VPN methods END */
+
+    void fetchCreditCardInfo(CommandFetchCreditCardCompletion completion);
+    void setProFlexi(bool newProFlexi);
 };
 
 } // namespace
