@@ -87,7 +87,7 @@ struct MEGA_API PublicLink
     string mAuthKey;
 
     PublicLink(handle ph, m_time_t cts, m_time_t ets, bool takendown, const char *authKey = nullptr);
-    PublicLink(PublicLink *plink);
+    PublicLink(const PublicLink& plink) = default;
 
     bool isExpired();
 };
@@ -176,14 +176,15 @@ struct MEGA_API Node : public NodeCore, FileFingerprint
     // display path from its root in the cloud (UTF-8)
     string displaypath() const;
 
-    // return mimetype type
-    MimeType_t getMimeType(bool checkPreview = false) const;
+    // match mimetype type
+    // checkPreview flag is only compatible with MimeType_t::MIME_TYPE_PHOTO
+    bool isIncludedForMimetype(MimeType_t mimetype, bool checkPreview = false) const;
 
     // node attributes
     AttrMap attrs;
 
     static const vector<string> attributesToCopyIntoPreviousVersions;
-    
+
     // 'sen' attribute
     bool isMarkedSensitive() const;
     bool isSensitiveInherited() const;
@@ -210,19 +211,21 @@ struct MEGA_API Node : public NodeCore, FileFingerprint
     static byte* decryptattr(SymmCipher*, const char*, size_t);
 
     // parse node attributes from an incoming buffer, this function must be called after call decryptattr
-    static void parseattr(byte*, AttrMap&, m_off_t, m_time_t&, string&, string&, FileFingerprint&);
+    // fingerprint output param is a raw fingerprint (i.e. without App prefixes)
+    static void parseattr(byte* bufattr, AttrMap& attrs, m_off_t size, m_time_t& mtime, string& fileName,
+                          string& fingerprint, FileFingerprint& ffp);
 
     // inbound share
-    Share* inshare = nullptr;
+    unique_ptr<Share> inshare;
 
     // outbound shares by user
-    share_map* outshares = nullptr;
+    unique_ptr<share_map> outshares;
 
     // outbound pending shares
-    share_map* pendingshares = nullptr;
+    unique_ptr<share_map> pendingshares;
 
     // incoming/outgoing share key
-    SymmCipher* sharekey = nullptr;
+    unique_ptr<SymmCipher> sharekey;
 
     // app-private pointer
     void* appdata = nullptr;
@@ -257,18 +260,17 @@ struct MEGA_API Node : public NodeCore, FileFingerprint
 
     } changed;
 
-    void setkey(const byte* = NULL);
 
+    void setKey(const string& key);
+    void setkey(const byte*);
     void setkeyfromjson(const char*);
 
-    void setUndecryptedKey(const std::string &undecryptedKey);
-
     void setfingerprint();
 
     void faspec(string*);
 
     NodeCounter getCounter() const;
-    void setCounter(const NodeCounter &counter, bool notify);
+    void setCounter(const NodeCounter &counter);  // to only be called by mNodeManger::setNodeCounter
 
     // parent
     // nullptr if is root node or top node of an inshare
@@ -305,11 +307,11 @@ struct MEGA_API Node : public NodeCore, FileFingerprint
     bool isbelow(NodeHandle) const;
 
     // handle of public link for the node
-    PublicLink* plink = nullptr;
+    unique_ptr<PublicLink> plink;
 
     void setpubliclink(handle, m_time_t, m_time_t, bool, const string &authKey = {});
 
-    bool serialize(string*) override;
+    bool serialize(string*) const override;
     static Node* unserialize(MegaClient& client, const string*, bool fromOldCache, std::list<std::unique_ptr<NewShare>>& ownNewshares);
 
     Node(MegaClient&, NodeHandle, NodeHandle, nodetype_t, m_off_t, handle, const char*, m_time_t);
@@ -345,7 +347,7 @@ struct MEGA_API Node : public NodeCore, FileFingerprint
         FLAGS_SIZE = 3
     };
 
-    typedef std::bitset<FLAGS_SIZE> Flags; 
+    typedef std::bitset<FLAGS_SIZE> Flags;
 
     // check if any of the flags are set in any of the anesestors
     bool anyExcludeRecursiveFlag(Flags excludeRecursiveFlags) const;
@@ -361,6 +363,22 @@ struct MEGA_API Node : public NodeCore, FileFingerprint
 
     static uint64_t getDBFlags(uint64_t oldFlags, bool isInRubbish, bool isVersion, bool isSensitive);
 
+    static bool getExtension(std::string& ext, const std::string& nodeName);
+    static bool isPhoto(const std::string& ext);
+    static bool isVideo(const std::string& ext);
+    static bool isAudio(const std::string& ext);
+    static bool isDocument(const std::string& ext);
+    static bool isSpreadsheet(const std::string& ext);
+    static bool isPdf(const std::string& ext);
+    static bool isPresentation(const std::string& ext);
+    static bool isArchive(const std::string& ext);
+    static bool isProgram(const std::string& ext);
+    static bool isMiscellaneous(const std::string& ext);
+    static bool isOfMimetype(MimeType_t mimetype, const std::string& ext);
+
+    bool isPhotoWithFileAttributes(bool checkPreview) const;
+    bool isVideoWithFileAttributes() const;
+
 private:
     // full folder/file key, symmetrically or asymmetrically encrypted
     // node crypto keys (raw or cooked -
@@ -370,12 +388,6 @@ private:
     // keeps track of counts of files, folder, versions, storage and version's storage
     NodeCounter mCounter;
 
-    bool getExtension(std::string& ext) const;
-    bool isPhoto(const std::string& ext, bool checkPreview) const;
-    bool isVideo(const std::string& ext) const;
-    bool isAudio(const std::string& ext) const;
-    bool isDocument(const std::string& ext) const;
-
     static nameid getExtensionNameId(const std::string& ext);
 };
 
@@ -510,7 +522,7 @@ struct MEGA_API LocalNode : public File
     LocalNode(Sync*);
     void init(nodetype_t, LocalNode*, const LocalPath&, std::unique_ptr<LocalPath>);
 
-    bool serialize(string*) override;
+    bool serialize(string*) const override;
     static LocalNode* unserialize( Sync* sync, const string* sData );
 
     ~LocalNode();
