@@ -52,10 +52,17 @@ typedef char __static_check_01__[sizeof(bool) == sizeof(char) ? 1 : -1];
 // signed 64-bit generic offset
 typedef int64_t m_off_t;
 
-// opaque filesystem fingerprint
-typedef uint64_t fsfp_t;
-
 namespace mega {
+
+// opaque filesystem fingerprint (eg on windows, volume serial number).  Let's make it a proper type so the compiler helps us not mix it up with other IDs
+struct fsfp_t
+{
+    uint64_t id = 0;
+    fsfp_t() {}
+    fsfp_t(uint64_t i) : id(i) {}
+    operator bool() { return id != 0; }
+};
+
 // within ::mega namespace, byte is unsigned char (avoids ambiguity when std::byte from c++17 and perhaps other defined ::byte are available)
 #if defined(USE_CRYPTOPP) && (CRYPTOPP_VERSION >= 600) && ((__cplusplus >= 201103L) || (__RPCNDR_H_VERSION__ == 500))
 using byte = CryptoPP::byte;
@@ -77,6 +84,7 @@ typedef unsigned char byte;
 #include <string>
 #include <chrono>
 #include <mutex>
+#include <thread>
 
 namespace mega {
 
@@ -152,6 +160,7 @@ class AuthRing;
 // Utils.h has functions m_time() and so on corresponding to time() which help us to use this type and avoid arithmetic overflow when working with time_t on systems where it's 32-bit
 typedef int64_t m_time_t;
 constexpr m_time_t mega_invalid_timestamp = 0;
+inline bool isValidTimeStamp(m_time_t t) { return t != mega_invalid_timestamp; }
 
 // monotonously increasing time in deciseconds
 typedef uint32_t dstime;
@@ -216,6 +225,7 @@ typedef enum ErrorCodes : int
     API_EBUSINESSPASTDUE = -28,     ///< Business account expired
     API_EPAYWALL = -29,             ///< Over Disk Quota Paywall
     LOCAL_ENOSPC = -1000,           ///< Insufficient space
+    LOCAL_ETIMEOUT = -1001,         ///< A request timed out.
 } error;
 
 class Error
@@ -346,9 +356,15 @@ typedef enum {
     FOLDERNODE,      // FOLDER - regular folder nodes
     ROOTNODE,        // ROOT - the cloud drive root node
     VAULTNODE,       // VAULT - vault, for "My backups" and other special folders
-    RUBBISHNODE      // RUBBISH - rubbish bin
+    RUBBISHNODE,     // RUBBISH - rubbish bin
 } nodetype_t;
 
+enum class TypeOfLink {
+    FOLDER,
+    FILE,
+    SET,
+};
+
 typedef enum { NO_SHARES = 0x00, IN_SHARES = 0x01, OUT_SHARES = 0x02, PENDING_OUTSHARES = 0x04, LINK = 0x08} ShareType_t;
 
 // MimeType_t maps to file extensionse declared at Node
@@ -356,7 +372,13 @@ typedef enum { MIME_TYPE_UNKNOWN    = 0,
                MIME_TYPE_PHOTO      = 1,    // photoExtensions, photoRawExtensions, photoImageDefExtension
                MIME_TYPE_AUDIO      = 2,    // audioExtensions longAudioExtension
                MIME_TYPE_VIDEO      = 3,    // videoExtensions
-               MIME_TYPE_DOCUMENT   = 4     // documentExtensions
+               MIME_TYPE_DOCUMENT   = 4,    // documentExtensions
+               MIME_TYPE_PDF        = 5,    // pdfExtensions
+               MIME_TYPE_PRESENTATION = 6,  // presentationExtensions
+               MIME_TYPE_ARCHIVE    = 7,    // archiveExtensions
+               MIME_TYPE_PROGRAM    = 8,    // programExtensions
+               MIME_TYPE_MISC       = 9,    // miscExtensions
+               MIME_TYPE_SPREADSHEET = 10,  // spreadsheetExtensions
              } MimeType_t;
 
 typedef enum { LBL_UNKNOWN = 0, LBL_RED = 1, LBL_ORANGE = 2, LBL_YELLOW = 3, LBL_GREEN = 4,
@@ -365,6 +387,7 @@ typedef enum { LBL_UNKNOWN = 0, LBL_RED = 1, LBL_ORANGE = 2, LBL_YELLOW = 3, LBL
 // node type key lengths
 const int FILENODEKEYLENGTH = 32;
 const int FOLDERNODEKEYLENGTH = 16;
+const int SETNODEKEYLENGTH = SymmCipher::KEYLENGTH;
 
 // Max nodes per putnodes command
 const unsigned MAXNODESUPLOAD = 1000;
@@ -393,7 +416,7 @@ class Cacheable
 public:
     virtual ~Cacheable() = default;
 
-    virtual bool serialize(string*) = 0;
+    virtual bool serialize(string*) const = 0;
 
     uint32_t dbid = 0;
     bool notified = false;
@@ -504,6 +527,10 @@ enum SyncError {
     UNABLE_TO_OPEN_DATABASE = 41,           // Unable to open state cache database.
     INSUFFICIENT_DISK_SPACE = 42,           // Insufficient space for download.
     FAILURE_ACCESSING_PERSISTENT_STORAGE = 43, // Failure accessing to persistent storage
+    MISMATCH_OF_ROOT_FSID = 44,             // The sync root's FSID changed.  So this is a different folder.  And, we can't identify the old sync db as the name depends on this
+    FILESYSTEM_FILE_IDS_ARE_UNSTABLE = 45,  // On MAC in particular, the FSID of a file in an exFAT drive can and does change spontaneously and frequently
+    FILESYSTEM_ID_UNAVAILABLE = 46,         // If we can't get a filesystem's id
+    UNABLE_TO_RETRIEVE_DEVICE_ID = 47,      // Unable to retrieve the ID of current device
 };
 
 enum SyncWarning {
@@ -629,7 +656,7 @@ typedef map<int, GenericHttpReq*> pendinghttp_map;
 typedef map<NodeHandle, unique_ptr<Node>> node_map;
 
 // maps node handles to Share pointers
-typedef map<handle, struct Share*> share_map;
+typedef map<handle, unique_ptr<struct Share>> share_map;
 
 // maps node handles NewShare pointers
 typedef list<struct NewShare*> newshare_list;
@@ -679,7 +706,7 @@ typedef enum { TREESTATE_NONE = 0,
                TREESTATE_SYNCED,
                TREESTATE_PENDING,
                TREESTATE_SYNCING,
-			   TREESTATE_IGNORED,
+               TREESTATE_IGNORED,
                } treestate_t;
 
 typedef enum { TRANSFERSTATE_NONE = 0, TRANSFERSTATE_QUEUED, TRANSFERSTATE_ACTIVE, TRANSFERSTATE_PAUSED,
@@ -736,6 +763,8 @@ typedef enum {
     //ATTR_DRIVE_NAMES = 35,                // (merged with ATTR_DEVICE_NAMES and removed) private - byte array - versioned
     ATTR_NO_CALLKIT = 36,                   // private, non-encrypted - char array in B64 - non-versioned
     ATTR_KEYS = 37,                         // private, non-encrypted (but encrypted to derived key from MK) - binary blob, non-versioned
+    ATTR_APPS_PREFS = 38,                   // private - byte array - versioned (apps preferences)
+    ATTR_CC_PREFS   = 39,                   // private - byte array - versioned (content consumption preferences)
 
 } attr_t;
 typedef map<attr_t, string> userattr_map;
@@ -758,7 +787,20 @@ typedef enum { RECOVER_WITH_MASTERKEY = 9, RECOVER_WITHOUT_MASTERKEY = 10, CANCE
 
 typedef enum { EMAIL_REMOVED = 0, EMAIL_PENDING_REMOVED = 1, EMAIL_PENDING_ADDED = 2, EMAIL_FULLY_ACCEPTED = 3 } emailstatus_t;
 
-typedef enum { RETRY_NONE = 0, RETRY_CONNECTIVITY = 1, RETRY_SERVERS_BUSY = 2, RETRY_API_LOCK = 3, RETRY_RATE_LIMIT = 4, RETRY_LOCAL_LOCK = 5, RETRY_UNKNOWN = 6} retryreason_t;
+#define DEFINE_RETRY_REASONS(expander) \
+    expander(0, RETRY_NONE) \
+    expander(1, RETRY_CONNECTIVITY) \
+    expander(2, RETRY_SERVERS_BUSY) \
+    expander(3, RETRY_API_LOCK) \
+    expander(4, RETRY_RATE_LIMIT) \
+    expander(5, RETRY_LOCAL_LOCK) \
+    expander(6, RETRY_UNKNOWN)
+
+typedef enum {
+#define DEFINE_RETRY_REASON(index, name) name = index,
+    DEFINE_RETRY_REASONS(DEFINE_RETRY_REASON)
+#undef DEFINE_RETRY_REASON
+} retryreason_t;
 
 typedef enum {
     STORAGE_UNKNOWN = -9,
@@ -808,6 +850,9 @@ typedef enum {
     ACCOUNT_TYPE_PROII = 2,
     ACCOUNT_TYPE_PROIII = 3,
     ACCOUNT_TYPE_LITE = 4,
+    ACCOUNT_TYPE_STARTER = 11,
+    ACCOUNT_TYPE_BASIC = 12,
+    ACCOUNT_TYPE_ESSENTIAL = 13,
     ACCOUNT_TYPE_BUSINESS = 100,
     ACCOUNT_TYPE_PRO_FLEXI = 101
 } AccountType;
@@ -836,6 +881,7 @@ typedef enum {
     REASON_ERROR_UNSERIALIZE_NODE   = 1,
     REASON_ERROR_DB_IO              = 2,
     REASON_ERROR_DB_FULL            = 3,
+    REASON_ERROR_DB_INDEX_OVERFLOW  = 4,
 } ErrorReason;
 
 // inside 'mega' namespace, since use C++11 and can't rely on C++14 yet, provide make_unique for the most common case.
@@ -967,7 +1013,7 @@ public:
     CacheableStatus(Type type, int64_t value);
 
     // serializes the object to a string
-    bool serialize(string* data) override;
+    bool serialize(string* data) const override;
 
     // deserializes the string to a SyncConfig object. Returns null in case of failure
     // returns a pointer to the unserialized value, owned by MegaClient passed as parameter
@@ -982,9 +1028,6 @@ public:
 
 private:
 
-    // need this to ensure serialization doesn't mutate state (Cacheable::serialize is non-const)
-    bool serialize(std::string& data) const;
-
     Type mType = STATUS_UNKNOWN;
     int64_t mValue = 0;
 
@@ -1178,8 +1221,19 @@ public:
 typedef std::map<NodeHandle, Node*> nodePtr_map;
 
 #ifdef ENABLE_CHAT
+
+class ScheduledFlags;
+class ScheduledMeeting;
+class ScheduledRules;
+class TextChat;
+
+using textchat_map = map<handle, TextChat*>;
+using textchat_vector = vector<TextChat*>;
+
 static constexpr int sfu_invalid_id = -1;
-#endif
+
+#endif // ENABLE_CHAT
+
 } // namespace mega
 
 #define MEGA_DISABLE_COPY(class_name) \
@@ -1206,4 +1260,207 @@ static constexpr int sfu_invalid_id = -1;
     MEGA_DEFAULT_COPY(class_name) \
     MEGA_DEFAULT_MOVE(class_name)
 
+typedef std::pair<std::string, std::string> StringPair;
+
+struct StringKeyPair
+{
+    std::string privKey;
+    std::string pubKey;
+
+    StringKeyPair(std::string&& privKey, std::string&& pubKey)
+    : privKey(std::move(privKey)), pubKey(std::move(pubKey))
+    {}
+};
+
+// A simple busy-wait lock for lightweight mutual exclusion.
+class Spinlock
+{
+    // Is the spinlock currently locked?
+    std::atomic_flag mLocked;
+
+public:
+    Spinlock()
+      : mLocked()
+    {
+        // Necessary until C++20.
+        mLocked.clear();
+    }
+
+    Spinlock(const Spinlock& other) = delete;
+
+    Spinlock& operator=(const Spinlock& rhs) = delete;
+
+    // Acquire exclusive ownership of this lock.
+    void lock()
+    {
+        // Poll until the loop is acquired.
+        while (!try_lock())
+            ;
+    }
+
+    // Try and acquire exclusive ownership of this lock.
+    bool try_lock()
+    {
+        return !mLocked.test_and_set();
+    }
+
+    // Release exclusive ownership of this lock.
+    void unlock()
+    {
+        mLocked.clear();
+    }
+}; // Spinlock
+
+namespace detail
+{
+
+template<typename T>
+struct IsRecursiveMutex
+  : std::false_type
+{
+}; // IsRecursiveMutex<T>
+
+template<>
+struct IsRecursiveMutex<std::recursive_mutex>
+  : std::true_type
+{
+}; // IsRecursiveMutex<std::recursive_mutex>
+
+template<typename T, bool IsRecursive = IsRecursiveMutex<T>::value>
+class CheckableMutex
+{
+    T mMutex;
+    std::atomic<std::thread::id> mOwner;
+
+public:
+    CheckableMutex()
+      : mMutex()
+      , mOwner(std::thread::id())
+    {
+    }
+
+    CheckableMutex(const CheckableMutex& other) = delete;
+
+    CheckableMutex& operator=(const CheckableMutex& rhs) = delete;
+
+    void lock()
+    {
+        auto id = std::this_thread::get_id();
+
+        assert(mOwner != id);
+
+        mMutex.lock();
+        mOwner = id;
+    }
+
+    bool owns_lock() const
+    {
+        return mOwner == std::this_thread::get_id();
+    }
+
+    bool try_lock()
+    {
+        auto id = std::this_thread::get_id();
+
+        if (mMutex.try_lock())
+        {
+            mOwner = id;
+            return true;
+        }
+
+        return false;
+    }
+
+    void unlock()
+    {
+        auto id = std::this_thread::get_id();
+
+        assert(mOwner == id);
+
+        mOwner = std::thread::id();
+        mMutex.unlock();
+
+        static_cast<void>(id);
+    }
+}; // CheckableMutex<T, false>
+
+template<typename T>
+class CheckableMutex<T, true>
+{
+    std::uint32_t mCount;
+    mutable Spinlock mLock;
+    T mMutex;
+    std::thread::id mOwner;
+
+public:
+    CheckableMutex()
+      : mCount(0)
+      , mLock()
+      , mMutex()
+      , mOwner()
+    {
+    }
+
+    CheckableMutex(const CheckableMutex& other) = delete;
+
+    CheckableMutex& operator=(const CheckableMutex& rhs) = delete;
+
+    void lock()
+    {
+        auto id = std::this_thread::get_id();
+
+        mMutex.lock();
+
+        std::lock_guard<Spinlock> guard(mLock);
+
+        mCount = mCount + 1;
+        mOwner = id;
+    }
+
+    bool owns_lock() const
+    {
+        auto id = std::this_thread::get_id();
+
+        std::lock_guard<Spinlock> guard(mLock);
+
+        return mOwner == id && mCount > 0;
+    }
+
+    bool try_lock()
+    {
+        auto id = std::this_thread::get_id();
+
+        if (!mMutex.try_lock())
+            return false;
+
+        std::lock_guard<Spinlock> guard(mLock);
+
+        mCount = mCount + 1;
+        mOwner = id;
+
+        return true;
+    }
+
+    void unlock()
+    {
+        auto id = std::this_thread::get_id();
+
+        std::lock_guard<Spinlock> guard(mLock);
+
+        assert(mCount);
+        assert(mOwner == id);
+
+        if (!--mCount)
+            mOwner = std::thread::id();
+
+        mMutex.unlock();
+
+        static_cast<void>(id);
+    }
+}; // CheckableMutex<T, true>
+
+} // detail
+
+using detail::CheckableMutex;
+
 #endif
