@@ -22,6 +22,11 @@
 #ifndef MEGAUSERNOTIFICATIONS_H
 #define MEGAUSERNOTIFICATIONS_H 1
 
+#include "json.h"
+#include "utils.h"
+
+#include <bitset>
+
 namespace mega {
 
 struct UserAlertRaw
@@ -139,7 +144,7 @@ namespace UserAlert
             bool seen = false;
         } pst;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static unique_ptr<Persistent> readBase(CacheableReader& r);
         static unique_ptr<Persistent> unserialize(string*);
         friend Base* unserializeNewUpdSched(string*, unsigned id);
@@ -160,9 +165,9 @@ namespace UserAlert
 
         void initTs(m_time_t dts, m_time_t rts);
 
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static IncomingPendingContact* unserialize(string*, unsigned id);
     };
 
@@ -172,10 +177,10 @@ namespace UserAlert
 
         ContactChange(UserAlertRaw& un, unsigned int id);
         ContactChange(int c, handle uh, const string& email, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
-        virtual bool checkprovisional(handle ou, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
+        virtual bool checkprovisional(handle ou, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static ContactChange* unserialize(string*, unsigned id);
     };
 
@@ -185,9 +190,9 @@ namespace UserAlert
 
         UpdatedPendingContactIncoming(UserAlertRaw& un, unsigned int id);
         UpdatedPendingContactIncoming(int s, handle uh, const string& email, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static UpdatedPendingContactIncoming* unserialize(string*, unsigned id);
     };
 
@@ -197,9 +202,9 @@ namespace UserAlert
 
         UpdatedPendingContactOutgoing(UserAlertRaw& un, unsigned int id);
         UpdatedPendingContactOutgoing(int s, handle uh, const string& email, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static UpdatedPendingContactOutgoing* unserialize(string*, unsigned id);
     };
 
@@ -209,9 +214,9 @@ namespace UserAlert
 
         NewShare(UserAlertRaw& un, unsigned int id);
         NewShare(handle h, handle uh, const string& email, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static NewShare* unserialize(string*, unsigned id);
     };
 
@@ -224,10 +229,10 @@ namespace UserAlert
 
         DeletedShare(UserAlertRaw& un, unsigned int id);
         DeletedShare(handle uh, const string& email, handle removerhandle, handle folderhandle, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
-        virtual void updateEmail(MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
+        virtual void updateEmail(MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static DeletedShare* unserialize(string*, unsigned id);
     };
 
@@ -241,9 +246,9 @@ namespace UserAlert
         NewSharedNodes(handle uh, handle ph, m_time_t timestamp, unsigned int id,
                        vector<handle>&& fileHandles, vector<handle>&& folderHandles);
 
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static NewSharedNodes* unserialize(string*, unsigned id);
     };
 
@@ -255,9 +260,9 @@ namespace UserAlert
         RemovedSharedNode(handle uh, m_time_t timestamp, unsigned int id,
                           vector<handle>&& handles);
 
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static RemovedSharedNode* unserialize(string*, unsigned id);
     };
 
@@ -268,9 +273,9 @@ namespace UserAlert
         UpdatedSharedNode(UserAlertRaw& un, unsigned int id);
         UpdatedSharedNode(handle uh, m_time_t timestamp, unsigned int id,
                           vector<handle>&& handles);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static UpdatedSharedNode* unserialize(string*, unsigned id);
     };
 
@@ -281,10 +286,10 @@ namespace UserAlert
 
         Payment(UserAlertRaw& un, unsigned int id);
         Payment(bool s, int plan, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
         string getProPlanName();
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static Payment* unserialize(string*, unsigned id);
     };
 
@@ -293,9 +298,9 @@ namespace UserAlert
         m_time_t expiryTime;
         PaymentReminder(UserAlertRaw& un, unsigned int id);
         PaymentReminder(m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static PaymentReminder* unserialize(string*, unsigned id);
     };
 
@@ -307,9 +312,9 @@ namespace UserAlert
 
         Takedown(UserAlertRaw& un, unsigned int id);
         Takedown(bool down, bool reinstate, int t, handle nh, m_time_t timestamp, unsigned int id);
-        virtual void text(string& header, string& title, MegaClient* mc);
+        virtual void text(string& header, string& title, MegaClient* mc) override;
 
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static Takedown* unserialize(string*, unsigned id);
     };
 
@@ -331,7 +336,7 @@ namespace UserAlert
             {}
 
         virtual void text(string& header, string& title, MegaClient* mc) override;
-        bool serialize(string* d) override;
+        bool serialize(string* d) const override;
         static NewScheduledMeeting* unserialize(string*, unsigned id);
     };
 
@@ -342,15 +347,15 @@ namespace UserAlert
         public:
             enum
             {
-                CHANGE_TYPE_TITLE,
-                CHANGE_TYPE_DESCRIPTION,
-                CHANGE_TYPE_CANCELLED,
-                CHANGE_TYPE_TIMEZONE,
-                CHANGE_TYPE_STARTDATE,
-                CHANGE_TYPE_ENDDATE,
-                CHANGE_TYPE_RULES,
+                CHANGE_TYPE_TITLE       = 0x01,
+                CHANGE_TYPE_DESCRIPTION = 0x02,
+                CHANGE_TYPE_CANCELLED   = 0x04,
+                CHANGE_TYPE_TIMEZONE    = 0x08,
+                CHANGE_TYPE_STARTDATE   = 0x10,
+                CHANGE_TYPE_ENDDATE     = 0x20,
+                CHANGE_TYPE_RULES       = 0x40,
 
-                CHANGE_TYPE_SIZE
+                CHANGE_TYPE_SIZE        = 7 // remember to update this when adding new values
             };
             struct StrChangeset { string oldValue, newValue; };
             struct TsChangeset  { m_time_t oldValue, newValue; };
@@ -359,16 +364,14 @@ namespace UserAlert
             const StrChangeset* getUpdatedTimeZone() const      { return mUpdatedTimeZone.get(); }
             const TsChangeset* getUpdatedStartDateTime() const  { return mUpdatedStartDateTime.get(); }
             const TsChangeset* getUpdatedEndDateTime() const    { return mUpdatedEndDateTime.get(); }
-            unsigned long getChanges() const                    { return mUpdatedFields.to_ulong(); }
-            bool hasChanged(int changeType) const
+            uint64_t getChanges() const                         { return mUpdatedFields.to_ullong(); }
+            bool hasChanged(uint64_t changeType) const
             {
-                return isValidChange(changeType)
-                        ? mUpdatedFields[changeType]
-                        : false;
+                return getChanges() & changeType;
             }
 
-            string changeToString(int changeType) const;
-            void addChange(int changeType, StrChangeset* = nullptr, TsChangeset* = nullptr);
+            string changeToString(uint64_t changeType) const;
+            void addChange(uint64_t changeType, StrChangeset* = nullptr, TsChangeset* = nullptr);
             Changeset() = default;
             Changeset(const std::bitset<CHANGE_TYPE_SIZE>& _bs,
                       unique_ptr<StrChangeset>& _titleCS,
@@ -393,16 +396,12 @@ namespace UserAlert
              */
             bool invariant() const
             {
+                const auto changes = getChanges();
                 return (mUpdatedFields.size() == CHANGE_TYPE_SIZE
-                        && (!mUpdatedFields[CHANGE_TYPE_TITLE]     || !!mUpdatedTitle)
-                        && (!mUpdatedFields[CHANGE_TYPE_TIMEZONE]  || !!mUpdatedTimeZone)
-                        && (!mUpdatedFields[CHANGE_TYPE_STARTDATE] || !!mUpdatedStartDateTime)
-                        && (!mUpdatedFields[CHANGE_TYPE_ENDDATE]   || !!mUpdatedEndDateTime));
-            }
-
-            bool isValidChange(int changeType) const
-            {
-                return static_cast<unsigned>(changeType) < static_cast<unsigned>(CHANGE_TYPE_SIZE);
+                        && (!(changes & CHANGE_TYPE_TITLE)     || mUpdatedTitle)
+                        && (!(changes & CHANGE_TYPE_TIMEZONE)  || mUpdatedTimeZone)
+                        && (!(changes & CHANGE_TYPE_STARTDATE) || mUpdatedStartDateTime)
+                        && (!(changes & CHANGE_TYPE_ENDDATE)   || mUpdatedEndDateTime));
             }
 
             void replaceCurrent(const Changeset& src)
@@ -450,7 +449,7 @@ namespace UserAlert
             {}
 
         virtual void text(string& header, string& title, MegaClient* mc) override;
-        bool serialize(string*) override;
+        bool serialize(string*) const override;
         static UpdatedScheduledMeeting* unserialize(string*, unsigned id);
     };
 
@@ -466,7 +465,7 @@ namespace UserAlert
             {}
 
         virtual void text(string& header, string& title, MegaClient* mc) override;
-        bool serialize(string* d) override;
+        bool serialize(string* d) const override;
         static DeletedScheduledMeeting* unserialize(string*, unsigned id);
     };
 #endif
@@ -531,8 +530,15 @@ private:
             std::transform(alertTypePerFolderNode.begin(), alertTypePerFolderNode.end(), std::back_inserter(v), [](const pair<handle, nameid>& p) { return p.first; });
             return v;
         }
+
+        bool areNodeVersions() const { return mAreNodeVersions; }
+        void areNodeVersions(const bool theyAre) { mAreNodeVersions = areNodeVersions() || theyAre; }
+
+        void squash(const ff& rhs);
+    private:
+	bool mAreNodeVersions = false;
     };
-    using notedShNodesMap = map<pair<handle, handle>, ff>;
+    using notedShNodesMap = map<pair<handle, handle>, ff>; // <<userhandle, parenthandle>,ff>
     notedShNodesMap notedSharedNodes;
     notedShNodesMap deletedSharedNodesStash;
     bool notingSharedNodes;
@@ -605,6 +611,7 @@ public:
     bool unserializeAlert(string* d, uint32_t dbid);
 
     // stash removal-alert noted nodes
+    void purgeNodeVersionsFromStash();
     void convertStashedDeletedSharedNodes();
     bool isDeletedSharedNodesStashEmpty() const;
     void stashDeletedNotedSharedNodes(handle originatingUser);
