@@ -1,4 +1,4 @@
-/**
+﻿/**
  * @file mega/utils.h
  * @brief Mega SDK various utilities and helper classes
  *
@@ -28,21 +28,10 @@
 #include <mutex>
 
 #include "types.h"
-
 #undef SSIZE_MAX
 #include "mega/mega_utf8proc.h"
 #undef SSIZE_MAX
 
-// Needed for Windows Phone (MSVS 2013 - C++ version 9.8)
-#if defined(_WIN32) && _MSC_VER <= 1800 && __cplusplus < 201103L && !defined(_TIMESPEC_DEFINED) && ! __struct_timespec_defined
-struct timespec
-{
-    long long	tv_sec; 	/* seconds */
-    long        tv_nsec;	/* nanoseconds */
-};
-# define __struct_timespec_defined  1
-#endif
-
 namespace mega {
 // convert 1...8 character ID to int64 integer (endian agnostic)
 #define MAKENAMEID1(a) (nameid)(a)
@@ -57,6 +46,7 @@ namespace mega {
 std::string toNodeHandle(handle nodeHandle);
 std::string toNodeHandle(NodeHandle nodeHandle);
 std::string toHandle(handle h);
+std::pair<bool, TypeOfLink> toTypeOfLink (nodetype_t type);
 #define LOG_NODEHANDLE(x) toNodeHandle(x)
 #define LOG_HANDLE(x) toHandle(x)
 class SimpleLogger;
@@ -98,8 +88,9 @@ struct MEGA_API PaddedCBC
      * @param iv Optional initialisation vector for encryption. Will use a
      *     zero IV if not given. If `iv` is a zero length string, a new IV
      *     for encryption will be generated and available through the reference.
+     * @return true if encryption was successful.
      */
-    static void encrypt(PrnGen &rng, string* data, SymmCipher* key, string* iv = NULL);
+    static bool encrypt(PrnGen &rng, string* data, SymmCipher* key, string* iv = NULL);
 
     /**
      * @brief Decrypts a string and strips the padding.
@@ -111,7 +102,7 @@ struct MEGA_API PaddedCBC
      * @param key AES key for decryption.
      * @param iv Optional initialisation vector for encryption. Will use a
      *     zero IV if not given.
-     * @return Void.
+     * @return true if decryption was successful.
      */
     static bool decrypt(string* data, SymmCipher* key, string* iv = NULL);
 };
@@ -423,6 +414,21 @@ public:
 
     static std::string stringToHex(const std::string& input);
     static std::string hexToString(const std::string& input);
+    /**
+     * @brief Converts a hexadecimal string to a uint64_t value. The input string may or may not have the hex prefix "0x".
+     *
+     * @param input The hexadecimal string to be converted (ex: "78b1bbbda5f32526", "0x10FF, "0x0001")
+     * @return The uint64_t value corresponding to the input hexadecimal string.
+    */
+    static uint64_t hexStringToUint64(const std::string &input);
+    /**
+     * @brief Converts a 8-byte numeric value to a 16-character lowercase hexadecimal string, with zero-padding if necessary.
+     *
+     * @param input The uint64_t value to be converted to a hexadecimal string.
+     * @return A 16-character lowercase hexadecimal string representation of the input value (ex: "78b1bbbda5f32526").
+     *
+    */
+    static std::string uint64ToHexString(uint64_t input);
 
     static int32_t toLower(const int32_t c)
     {
@@ -462,6 +468,25 @@ public:
                                const std::string& search,
                                const std::string& replacement);
 
+    // join({"a", "new", "loom"}, "; ") -> "a; new; loom"
+    static std::string join(const std::vector<std::string>& items, const std::string& with);
+    static bool startswith(const std::string& str, const std::string& start);
+    static bool startswith(const std::string& str, char chr);
+    static bool endswith(const std::string& str, char chr);
+    static const std::string _trimDefaultChars;
+    // return string with trimchrs removed from front and back of given string str
+    static string trim(const string& str, const string& trimchars = _trimDefaultChars);
+
+
+    // --- environment functions that work with Unicode UTF-8 on Windows (set/unset/get) ---
+
+    static bool hasenv(const std::string& key);
+    // sets *out_found if found
+    static std::string getenv(const std::string& key, bool* out_found);
+    // return def if value not found
+    static std::string getenv(const std::string& key, const std::string& def);
+    static void setenv(const std::string& key, const std::string& value);
+    static void unsetenv(const std::string& key);
 };
 
 // for pre-c++11 where this version is not defined yet.
@@ -656,6 +681,10 @@ std::pair<bool, int64_t> generateMetaMac(SymmCipher &cipher, FileAccess &ifAcces
 
 std::pair<bool, int64_t> generateMetaMac(SymmCipher &cipher, InputStreamAccess &isAccess, const int64_t iv);
 
+bool CompareLocalFileMetaMacWithNodeKey(FileAccess* fa, const std::string& nodeKey, int type);
+
+bool CompareLocalFileMetaMacWithNode(FileAccess* fa, Node* node);
+
 // Helper class for MegaClient.  Suitable for expansion/templatizing for other use caes.
 // Maintains a small thread pool for executing independent operations such as encrypt/decrypt a block of data
 // The number of threads can be 0 (eg. for helper MegaApi that deals with public folder links) in which case something queued is
@@ -1013,8 +1042,35 @@ struct SyncTransferCounts
 };
 
 // creates a new id filling `id` with random bytes, up to `length`
-void resetId(char *id, size_t length, PrnGen& rng);
+void resetId(char* id, size_t length, PrnGen& rng);
+
+// write messsage and strerror(aerrno) to log as an error
+void reportError(const std::string& message, int aerrno = -1);
+
+#ifdef WIN32
+
+// as per (non C library standard) unix API
+inline void sleep(int sec) {
+    Sleep(sec * 1000);
+}
+
+// as per (non C library standard) unix API
+// sleep for given number of microseconds
+inline void usleep(int microsec) {
+    Sleep(microsec / 1000);
+}
+
+// print messgae: error-num: error-description
+void reportWindowsError(const std::string& message, DWORD error = 0xFFFFFFFF);
+
+#endif // WIN32
+
+// returns the direction type of a connection
+string connDirectionToStr(direction_t directionType);
+
+// Translate retry reason into a human-friendly string.
+const char* toString(retryreason_t reason);
 
 } // namespace mega
 
-#endif
+#endif // MEGA_UTILS_H
